-------------------------------------------------------------------------------
--
-- OPB Flash controller
--
--
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity opb_flash is
  
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector(0 to 31) := X"30000000";
    C_HIGHADDR   : std_logic_vector(0 to 31) := X"3FFFFFFF");

  port (
    OPB_Clk     : in  std_logic;
    OPB_Rst     : in  std_logic;
    OPB_ABus    : in  std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in  std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_DBus    : in  std_logic_vector(0 to C_OPB_DWIDTH-1);
    OPB_RNW     : in  std_logic;
    OPB_select  : in  std_logic;
    OPB_seqAddr : in  std_logic;        -- Sequential Address
    Sln_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sln_errAck  : out std_logic;        -- (unused)
    Sln_retry   : out std_logic;        -- (unused)
    Sln_toutSup : out std_logic;        -- Timeout suppress
    Sln_xferAck : out std_logic;        -- Transfer acknowledge
    
    PB_DIN   : in std_logic_vector(0 to 15);      -- FLASH data bus
    PB_DOUT  : out std_logic_vector(0 to 15);     -- FLASH data bus output
    PB_DOUT_EN : out std_logic;                   -- FLASH data bus output enable
    PB_A   : out std_logic_vector (0 to 19);      -- FLASH address bus
    PB_OE_N  : out std_logic;                       -- FLASH output enable
    PB_WE_N  : out std_logic;                       -- FLASH write enable
    FLASH_CE_N  : out std_logic                        -- FLASH byte enable 
    );       

end opb_flash;

architecture Behavioral of opb_flash is

  constant FLASH_AWIDTH : integer := 19;  -- Number of address lines on the FLASH
  constant FLASH_DWIDTH : integer := 8;  -- Number of data lines on the FLASH
    
  signal RNW : std_logic;
  signal FLASH_DI: std_logic_vector(0 to 31);
  signal ABus : std_logic_vector(0 to FLASH_AWIDTH-1);
  signal chip_select : std_logic;
  signal output_enable : std_logic;
  signal WE : std_logic;
  signal BE : std_logic_vector(0 to 3);
  signal A30A31: std_logic_vector (0 to 1);
  signal FLASH_RD_DATA : std_logic_vector (0 to 31);
  signal FLASH_WR_DATA : std_logic_vector (0 to 31);
  signal TOSUP : std_logic;
    
  signal CE : std_logic;           -- FLASH chip enalbe
  signal OE : std_logic;           -- FLASH output enalbe
  signal xfer_start: std_logic;    -- Transfer start
  signal xfer_end: std_logic;      -- Transfer end     
  signal xfer_idle: std_logic;     -- No transfer pending
  signal obuf_oe: std_logic;       -- Data output buffer enable
  signal xfer_done: std_logic;     -- Transfer has been done
  signal wait_expire: std_logic; 
  signal wait_counter: std_logic_vector ( 0 to 3) ;
  
  -- Critical: Sln_xferAck is generated directly from state bit 0!
  constant STATE_BITS : integer := 3;
  constant ST_IDLE      : std_logic_vector(0 to STATE_BITS-1) := "000";
  constant ST_SELECTED  : std_logic_vector(0 to STATE_BITS-1) := "001";
  constant ST_WAIT      : std_logic_vector(0 to STATE_BITS-1) := "011";
  constant ST_XFER      : std_logic_vector(0 to STATE_BITS-1) := "111";
  constant ST_HOLD      : std_logic_vector(0 to STATE_BITS-1) := "100";
  constant ST_WAITWORD  : std_logic_vector(0 to STATE_BITS-1) := "101";
  constant ST_XFERWORD  : std_logic_vector(0 to STATE_BITS-1) := "110";

  signal present_state, next_state : std_logic_vector(0 to STATE_BITS-1);

begin

  register_opb_inputs: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      FLASH_DI <= (others => '0');
      ABus <= (others => '0');
      RNW <= '0';
      A30A31 <= (others => '0');
      BE <= (others => '0');
    elsif OPB_Clk'event and OPB_Clk = '1' then
      FLASH_DI <= OPB_DBus(0 to C_OPB_DWIDTH-1);
      ABus <= OPB_ABus(C_OPB_AWIDTH-1-(FLASH_AWIDTH-1) to C_OPB_AWIDTH-1);
      RNW <= OPB_RNW;
      A30A31 <= OPB_ABus(30 to 31);
      BE <= OPB_BE(0 to 3);
    end if;
  end process register_opb_inputs;

  register_opb_outputs: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      Sln_DBus <= (others => '0');
    elsif OPB_Clk'event and OPB_Clk = '1' then
--     if (output_enable = '1' and RNW = '1')  then
     if (output_enable = '1' )  then
        Sln_DBus <= FLASH_RD_DATA;
     else
       Sln_DBus <= (others => '0');
     end if;
    end if;
  end process register_opb_outputs;

  -- Unused outputs
  Sln_errAck  <= '0';
  Sln_retry   <= '0';
  Sln_toutSup <= TOSUP;
  --Sln_DBus(FLASH_DWIDTH to C_OPB_DWIDTH-1) <= (others => '0');

  chip_select <=
    '1' when OPB_select = '1' and
        OPB_ABus(0 to C_OPB_AWIDTH-2-FLASH_AWIDTH) =
        C_BASEADDR(0 to C_OPB_AWIDTH-2-FLASH_AWIDTH) else
    '0';

  Sln_xferAck <= xfer_end;

  -- Sequential part of the FSM
  fsm_seq : process(OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      present_state <= ST_IDLE;
    elsif OPB_Clk'event and OPB_Clk = '1' then
      present_state <= next_state;
    end if;
  end process fsm_seq;

  -- Combinational part of the FSM
  fsm_comb : process(present_state, chip_select, OPB_Select , wait_expire)
  begin   
      case present_state is
        when ST_IDLE =>
          if chip_select = '1' then
            next_state <= ST_SELECTED;
          else
            next_state <= ST_IDLE;
          end if;
   
        when ST_SELECTED =>
          if OPB_Select = '1' then
              next_state <= ST_WAIT;
          else
            next_state <= ST_IDLE;
          end if;
   
        when ST_WAIT =>
          if OPB_Select = '1' then
             if ( wait_expire = '1' ) then   
                next_state <= ST_XFER;
             else 
                next_state <= ST_WAIT;
             end if;
          else
            next_state <= ST_IDLE;
          end if;
   
        -- State encoding is critical here: xfer must only be true here
        when ST_XFER =>
          next_state <= ST_IDLE;
          
        when others =>
          next_state <= ST_IDLE;
      end case;

  end process fsm_comb;

  fsm_output: process (present_state) 
  begin
     output_enable <= '0';
     xfer_start <= '0';
     xfer_end <= '0';   
     xfer_idle <= '0';
     case present_state is  
        when ST_IDLE =>
           xfer_idle <= '1';
        when ST_SELECTED =>
           xfer_start <= '1';
        when ST_WAIT =>      
           output_enable <= '1';
        when ST_XFER =>        
           xfer_end <= '1';
        when others => 
           output_enable <= '0';
           xfer_start <= '0';
           xfer_end <= '0';
           xfer_idle <= '0';
     end case;
  end process fsm_output;
  
  -- The following process generates the SFLASH OE output signal
  flash_oe_reg: process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then 
        OE <= '1';
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if ( chip_select = '1' and xfer_idle = '1') then 
           OE <= not OPB_RNW;
        elsif ( xfer_end = '1') then 
           OE <= '1';
        end if;
     end if;
  end process flash_oe_reg;

  -- The following process generates the flash chip enable output signal
  flash_ce_reg: process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then 
        CE <= '1';
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if ( chip_select = '1' and xfer_idle = '1') then 
           CE <= '0';
        elsif ( xfer_end = '1') then 
           CE <= '1';
        end if;
     end if;
  end process flash_ce_reg;                              


  process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then 
        TOSUP <= '0';
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if ( chip_select = '1' and xfer_idle = '1') then 
           TOSUP <= '1';
        elsif ( xfer_end = '1') then 
           TOSUP <= '0';
        end if;
     end if;
  end process;
    
  -- The following process generates the flash WE output signal
  flash_we_reg: process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then 
        WE <= '1';
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if ( xfer_start = '1') then 
           WE <= RNW;
        elsif (wait_expire = '1') then
           WE <= '1';
        end if;
     end if;
  end process flash_we_reg;                              
  
  -- Wait counter
  process (OPB_Clk, OPB_Rst) 
  begin
     if ( OPB_Rst = '1') then 
        wait_counter <= (others => '0');
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if ( xfer_start = '1' or xfer_idle = '1' ) then 
           wait_counter <= "0000";
        else 
           wait_counter <= wait_counter + '1';
        end if;
     end if;
  end process;
  
  process ( wait_counter, RNW) 
  begin
    if (RNW = '1') then
       if ( wait_counter = "0101"   ) then 
          wait_expire <= '1';
       else 
          wait_expire <= '0';
       end if;
    else 
       if ( wait_counter = "1011"   ) then 
          wait_expire <= '1';
       else 
          wait_expire <= '0';
       end if;
    end if;
  end process;
  
  --
  FLASH_WR_DATA <= FLASH_DI ;
  FLASH_RD_DATA <= PB_DIN(8 to 15) & PB_DIN(8 to 15) & PB_DIN(8 to 15) & PB_DIN(8 to 15);
  
  -- The following assign the flash output signals
  PB_OE_N <=  OE;
  PB_WE_N <= WE;
  FLASH_CE_N <= CE;
  PB_A <= '0' & ABus;
  PB_DOUT <= "00000000" & FLASH_WR_DATA(24 to 31);
  --FLASH_DO <= FLASH_RD_DATA;
  PB_DOUT_EN <= not CE and OE;


end Behavioral;


