-------------------------------------------------------------------------------
--
-- OPB SRAM controller
--
--
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;

entity opb_sram is

  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector(0 to 31) := X"20000000";
    C_HIGHADDR   : std_logic_vector(0 to 31) := X"2FFFFFFF");

  port (
    OPB_Clk     : in  std_logic;
    OPB_Rst     : in  std_logic;
    OPB_ABus    : in  std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in  std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_DBus    : in  std_logic_vector(0 to C_OPB_DWIDTH-1);
    OPB_RNW     : in  std_logic;
    OPB_select  : in  std_logic;
    OPB_seqAddr : in  std_logic;        -- Sequential Address
    Sln_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sln_errAck  : out std_logic;        -- (unused)
    Sln_retry   : out std_logic;        -- (unused)
    Sln_toutSup : out std_logic;        -- Timeout suppress
    Sln_xferAck : out std_logic;        -- Transfer acknowledge
    
    PB_DIN   : in std_logic_vector(0 to 15);      -- SRAM data bus
    PB_DOUT  : out std_logic_vector(0 to 15);     -- SRAM data bus output
    PB_DOUT_EN : out std_logic;                   -- SRAM data bus output enable
    PB_A   : out std_logic_vector (0 to 19);      -- SRAM address bus
    RAM_CE : out std_logic;                       -- SRAM chip select
    PB_OE_N : out std_logic;                      -- SRAM output enable
    PB_WE_N : out std_logic;                      -- SRAM write enable
    PB_UB_N : out std_logic;                      -- SRAM byte enable
    PB_LB_N : out std_logic                       -- SRAM byte enable 
    );       

end opb_sram;

architecture Behavioral of opb_sram is

  constant RAM_AWIDTH : integer := 18;  -- Number of address lines on the RAM
  constant RAM_DWIDTH : integer := 16;  -- Number of data lines on the RAM

  signal RNW : std_logic;
  signal RAM_DI: std_logic_vector(0 to 31);
  signal ABus : std_logic_vector(0 to RAM_AWIDTH-1);
  signal chip_select : std_logic;
  signal output_enable : std_logic;
  signal WE : std_logic;
  signal BE : std_logic_vector(0 to 3);
  signal A31: std_logic;
  signal A30R: std_logic;
  signal RAM_RD_DATA : std_logic_vector (0 to 31);
  signal RAM_RD_DATA_MUX : std_logic_vector (0 to 31);
  signal RAM_WR_DATA : std_logic_vector (0 to 31);
  
  signal CE : std_logic;           -- SRAM chip enalbe
  signal OE : std_logic;           -- SRAM output enalbe
  signal xfer_start: std_logic;    -- Transfer start
  signal xfer_end: std_logic;      -- Transfer end
  signal xfer_idle: std_logic;     -- No transfer pending
  signal obuf_oe: std_logic;       -- Data output buffer enable
  signal xfer_done: std_logic;     -- Transfer has been done
  signal byte_start: std_logic;
  signal word_address: std_logic;
  signal byte_mode: std_logic;
  signal halfword_mode: std_logic;
  signal word_mode: std_logic;
  signal A30 : std_logic;
  signal word_access: std_logic;   
  signal word_latch: std_logic;
  signal set_a30: std_logic;
  signal DIN_Latch: std_logic_vector(0 to 15);
  
  -- Critical: Sln_xferAck is generated directly from state bit 0!
  constant STATE_BITS : integer := 3;
  constant ST_IDLE      : std_logic_vector(0 to STATE_BITS-1) := "000";
  constant ST_SELECTED  : std_logic_vector(0 to STATE_BITS-1) := "001";
  constant ST_WAIT      : std_logic_vector(0 to STATE_BITS-1) := "011";
  constant ST_XFER      : std_logic_vector(0 to STATE_BITS-1) := "111";
  constant ST_WORDINIT  : std_logic_vector(0 to STATE_BITS-1) := "010";
  constant ST_HOLD      : std_logic_vector(0 to STATE_BITS-1) := "100";
  constant ST_WAITWORD  : std_logic_vector(0 to STATE_BITS-1) := "101";
  constant ST_XFERWORD  : std_logic_vector(0 to STATE_BITS-1) := "110";

  signal present_state, next_state : std_logic_vector(0 to STATE_BITS-1);

begin

  register_opb_inputs: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      RAM_DI <= (others => '0');
      ABus <= (others => '0');
      RNW <= '0';
      A31 <= '0';
      A30R <= '0';
      BE <= (others => '0');
    elsif OPB_Clk'event and OPB_Clk = '1' then
      RAM_DI <= OPB_DBus(0 to C_OPB_DWIDTH-1);
      if ( word_address = '1') then
         ABus <= OPB_ABus(C_OPB_AWIDTH-2-(RAM_AWIDTH-1) to C_OPB_AWIDTH-3) & '1';
      else
         ABus <= OPB_ABus(C_OPB_AWIDTH-2-(RAM_AWIDTH-1) to C_OPB_AWIDTH-2);
      end if ;
      A31 <= OPB_ABus(31);
      A30R <= OPB_ABus(30);
      if (xfer_idle = '1') then
         BE <= OPB_BE(0 to 3);
         RNW <= OPB_RNW;
      end if;
    end if;
  end process register_opb_inputs;

--  register_opb_outputs: process (OPB_Clk, OPB_Rst)
--  begin
--    if OPB_Rst = '1' then
--      Sln_DBus <= (others => '0');
--    elsif OPB_Clk'event and OPB_Clk = '1' then
----     if (output_enable = '1' and RNW = '1')  then
--     if (output_enable = '1')  then
--        Sln_DBus <= RAM_RD_DATA_MUX;
--     else
--       Sln_DBus <= (others => '0');
--     end if;
--    end if;
--  end process register_opb_outputs;

  process (RAM_RD_DATA_MUX, output_enable)
  begin
     if (output_enable = '1')  then
        Sln_DBus <= RAM_RD_DATA_MUX;
     else
        Sln_DBus <= (others => '0');
     end if;
  end process;

  -- Unused outputs
  Sln_errAck  <= '0';
  Sln_retry   <= '0';
  Sln_toutSup <= '0';
  --Sln_DBus(RAM_DWIDTH to C_OPB_DWIDTH-1) <= (others => '0');

  chip_select <=
    '1' when OPB_select = '1' and
        OPB_ABus(0 to C_OPB_AWIDTH-2-RAM_AWIDTH) =
        C_BASEADDR(0 to C_OPB_AWIDTH-2-RAM_AWIDTH) else
    '0';

  Sln_xferAck <= xfer_done;

  -- Sequential part of the FSM
  fsm_seq : process(OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      present_state <= ST_IDLE;
    elsif OPB_Clk'event and OPB_Clk = '1' then
      present_state <= next_state;
    end if;
  end process fsm_seq;

  -- Combinational part of the FSM
  fsm_comb : process(present_state, chip_select, OPB_Select, word_mode)
  begin   
      case present_state is
        when ST_IDLE =>
          if chip_select = '1' then
            next_state <= ST_SELECTED;
          else
            next_state <= ST_IDLE;
          end if;
   
        when ST_SELECTED =>
          if OPB_Select = '1' then
              next_state <= ST_XFER;
          else
              next_state <= ST_IDLE;
          end if;
   
        -- State encoding is critical here: xfer must only be true here
        when ST_XFER =>
          if word_mode = '1' then
              next_state <= ST_HOLD;
          else 
               next_state <= ST_IDLE;  
          end if;

        when ST_HOLD =>
          if OPB_Select = '1' then
              next_state <= ST_XFERWORD;
          else
            next_state <= ST_IDLE;
          end if;
   
        when ST_XFERWORD =>
            next_state <= ST_IDLE;
        when others =>
          next_state <= ST_IDLE;
      end case;

  end process fsm_comb;

  fsm_output: process (present_state, word_mode)
  begin
     output_enable <= '0';
     xfer_start <= '0';
     byte_start <= '0';
     xfer_end <= '0';
     xfer_idle <= '0';
     xfer_done <= '0';
     word_latch <= '0';
     word_address <= '0';
     set_a30 <= '0';
     case present_state is
        when ST_IDLE =>
           xfer_idle <= '1';
           byte_start <= '1';
        when ST_SELECTED =>
           xfer_start <= '1';
        when ST_XFER =>
           output_enable <= not word_mode;
           word_latch <= '1';
           xfer_end <= not word_mode;
           xfer_done <= not word_mode; 
           set_a30 <= word_mode;
           word_address <= word_mode;
        when ST_HOLD =>
           xfer_start <= '1';
           word_address <= '1';
        when ST_XFERWORD =>
           output_enable <= '1';
           xfer_end <= '1';
           xfer_done <= '1';

        when others =>
           output_enable <= '0';
           byte_start <= '0';
           xfer_start <= '0';
           xfer_end <= '0';
           xfer_idle <= '0';
           set_a30 <= '0';
           word_latch <= '0';
           word_address <= '0';
     end case;
  end process fsm_output;

  -- The following process generates the SRAM OE output signal
  sram_oe_reg: process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then 
        OE <= '1';
     elsif OPB_Clk 'event and OPB_Clk = '1' then
        if ( next_state = ST_SELECTED) then
           OE <= not OPB_RNW;
        elsif ( next_state = ST_IDLE) then
           OE <= '1';
        end if;
     end if;
  end process sram_oe_reg;

  -- The following process generates the SRAM chip enable output signal
  sram_ce_reg: process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then
        CE <= '1';
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if ( next_state = ST_SELECTED) then
           CE <= '0';
        elsif ( next_state = ST_IDLE) then
           CE <= '1';
        end if;
     end if;
  end process sram_ce_reg;                              
  
  -- The following process generates the SRAM WE output signal
  sram_we_reg: process (OPB_Clk, OPB_Rst)
  begin 
     if OPB_Rst = '1' then
        WE <= '1';
     elsif OPB_Clk 'event and OPB_Clk = '0' then
        if ( xfer_start = '1' and chip_select = '1' ) then 
           WE <= RNW;
        else 
           WE <= '1';
        end if;
     end if;
  end process sram_we_reg;
  
  -- Detect the current access mode
  process (BE) 
  begin
     byte_mode <= '0';
     halfword_mode <= '1';
     word_mode <= '0';
     case BE is
        when "0001" => 
           byte_mode <= '1';
           halfword_mode <= '0';
           word_mode <= '0';
        when "0010" =>
           byte_mode <= '1';
           halfword_mode <= '0';
           word_mode <= '0';
        when "0100" =>
           byte_mode <= '1';
           halfword_mode <= '0';
           word_mode <= '0';
        when "1000" =>
           byte_mode <= '1';
           halfword_mode <= '0';
           word_mode <= '0';
        when "0011" =>
           byte_mode <= '0';
           halfword_mode <= '1';
           word_mode <= '0';
        when "1100" =>
           byte_mode <= '0';
           halfword_mode <= '1';
           word_mode <= '0';
        when "1111" =>
           byte_mode <= '0';
           halfword_mode <= '0';
           word_mode <= '1';
        when others =>
           byte_mode <= '0';
           halfword_mode <= '1';
           word_mode <= '0';
   end case;
  end process;
  
  process ( OPB_Clk, OPB_Rst) 
  begin
     if (OPB_Rst = '1') then
        DIN_Latch <= (others => '0');
     elsif ( OPB_Clk 'event and OPB_Clk = '1') then 
        if ( word_latch = '1') then 
           DIN_Latch <= PB_DIN;
        end if;
     end if;
  end process;
  
  process (OPB_Clk, OPB_Rst) 
  begin
     if (OPB_Rst = '1') then 
        word_access <= '0';
     elsif (OPB_Clk 'event and OPB_Clk = '1') then
        if (byte_start = '1' or xfer_idle = '1' or xfer_end = '1' ) then
              word_access <= '0';
        elsif ( set_a30 = '1' and word_mode = '1') then
              word_access <= '1';
        end if;
     end if;
  end process;
  
  --
  
  --RAM_WR_DATA <= RAM_DI; 
  process ( RAM_DI, word_access, word_mode, A30R)
  begin
     if ( word_mode = '0') then 
        if ( A30R = '1') then 
           RAM_WR_DATA(16 to 31) <= RAM_DI(16 to 31);
        else 
           RAM_WR_DATA(16 to 31) <= RAM_DI(0 to 15);
        end if;
     else 
        if (word_access = '1') then 
           RAM_WR_DATA(16 to 31) <= RAM_DI(16 to 31);
        else 
           RAM_WR_DATA(16 to 31) <= RAM_DI(0 to 15);
        end if;
     end if;
  end process;
  
 --RAM_RD_DATA <= DIN_Latch & PB_DIN;

  process ( PB_DIN, DIN_Latch, A31, byte_mode) 
  begin
     if ( A31 = '0' and byte_mode = '1') then 
        RAM_RD_DATA(24 to 31) <= PB_DIN(0 to 7);
     else 
        RAM_RD_DATA(24 to 31) <= PB_DIN(8 to 15);     
     end if;
     
     RAM_RD_DATA(16 to 23) <= PB_DIN(0 to 7) ;
     RAM_RD_DATA(0 to 15) <= DIN_Latch ;
  end process;      
  
  process ( RAM_RD_DATA, byte_mode, halfword_mode) 
  begin
     if ( byte_mode = '1') then 
        RAM_RD_DATA_MUX <= RAM_RD_DATA(24 to 31) & RAM_RD_DATA(24 to 31) & RAM_RD_DATA(24 to 31) & RAM_RD_DATA(24 to 31);
     elsif (halfword_mode = '1' ) then 
        RAM_RD_DATA_MUX <= RAM_RD_DATA(16 to 31) & RAM_RD_DATA(16 to 31);
     else 
        RAM_RD_DATA_MUX <=  RAM_RD_DATA;
     end if;
  end process;
   
  
  ----
  
  A30 <= ABus(RAM_AWIDTH-1) or word_access;
  
  process (A31, CE, byte_mode, RNW) 
  begin                           
     if ( RNW = '0') then 
        if ( byte_mode = '1') then 
           if ( A31 = '1') then
              PB_UB_N <= '1';
              PB_LB_N <= CE;
           else
              PB_UB_N <= CE;
              PB_LB_N <= '1';
           end if;
        else 
           PB_UB_N <= CE;
           PB_LB_N <= CE;
        end if;
    else 
       PB_UB_N <= CE;
       PB_LB_N <= CE;
    end if;
  end process;
  
  -- The following assign the SRAM output signals
  PB_OE_N <= OE;
  PB_WE_N <= WE;
  RAM_CE <= CE;
  --PB_UB_N <= CE;
  --PB_LB_N <= CE;
  PB_A <= "00" & ABus(0 to RAM_AWIDTH-1);  -- & A30;
  PB_DOUT <= RAM_WR_DATA(16 to 31);
  PB_DOUT_EN <=  (not CE) and OE;


end Behavioral;

-- Local Variables:
-- compile-command: "ghdl -a opb_bram.vhd"
-- End:
